@@ -8,8 +8,8 @@ import pytorch3d.transforms.transform3d as pt3d  # type: ignore
 import torch
 import torch.nn.functional as F
 from einops import repeat  # type: ignore
-from pykdtree.kdtree import KDTree  # type: ignore
 from pytorch3d.ops import knn_points, sample_farthest_points  # type: ignore
+from pytorch3d.transforms.rotation_conversions import _axis_angle_rotation  # type: ignore
 from torch import Tensor
 
 
@@ -132,7 +132,7 @@ def get_o3d_pcd_from_tensor(
     if isinstance(pcd, Tensor):
         p = pcd.clone().detach().cpu().numpy()
     else:
-        p = np.copy(pcd)  # type: ignore
+        p = np.copy(pcd)
 
     pcd_o3d.points = o3d.utility.Vector3dVector(p[:, :3])
 
@@ -146,7 +146,7 @@ def get_o3d_pcd_from_tensor(
         if isinstance(colors, Tensor):
             c = colors.clone().detach().cpu().numpy()
         else:
-            c = np.copy(colors)  # type: ignore
+            c = np.copy(colors)
 
         pcd_o3d.colors = o3d.utility.Vector3dVector(c)
 
@@ -267,10 +267,12 @@ def random_point_sampling(pcd: Tensor, num_points: int) -> Tensor:
     batch_size, original_num_points, _ = pcd.shape
 
     weights = torch.ones((batch_size, original_num_points), dtype=torch.float)
+    weights = weights.to(pcd.device)
+
     replacement = original_num_points < num_points
     indices_to_sample = torch.multinomial(weights, num_points, replacement=replacement)
 
-    batch_indices = torch.arange(batch_size).view(batch_size, 1)
+    batch_indices = torch.arange(batch_size).reshape(batch_size, 1)
     sampled_points = pcd[batch_indices, indices_to_sample]
 
     if batched:
@@ -302,8 +304,8 @@ def get_distances_matrix(pcd_src: Tensor, pcd_trg: Tensor) -> Tensor:
     _, num_pts_trg, _ = pcd_trg.shape
 
     dot_src_trg = -2 * torch.matmul(pcd_src, pcd_trg.permute(0, 2, 1))
-    src_square = torch.sum(pcd_src ** 2, -1).view(size_batch, num_pts_src, 1)
-    trg_square = torch.sum(pcd_trg ** 2, -1).view(size_batch, 1, num_pts_trg)
+    src_square = torch.sum(pcd_src**2, -1).reshape(size_batch, num_pts_src, 1)
+    trg_square = torch.sum(pcd_trg**2, -1).reshape(size_batch, 1, num_pts_trg)
 
     mat_dist_squared = dot_src_trg + src_square + trg_square
     mat_dist_squared[mat_dist_squared < 0.0] = 0.0
@@ -548,6 +550,34 @@ def rotate_pcd(pcd: Tensor, rotation: Tensor) -> Tensor:
     return _apply_pt3d_transform(pcd, transform, transform_normals=True)
 
 
+def rotate_pcd_around_axis(pcd: Tensor, axis: str, angle: float) -> Tensor:
+    """Rotate a point cloud of the given angle around the given axis.
+
+    Args:
+        pcd: The point cloud(s) with shape ([B,] NUM_POINTS, D).
+        axis: The axis around which the point cloud(s) will be rotated.
+        angle: The required rotation in degrees.
+
+    Returns:
+        The rotated point cloud(s).
+    """
+    assert axis.upper() in ["X", "Y", "Z"]
+
+    batched, [pcd] = batchify([pcd], 3)
+
+    theta = torch.deg2rad(torch.tensor(angle))
+    # _axis_angle_rotation returns a rotation matrix thought to operate
+    # on column vectors, since we use row vectors we need to transpose
+    matrix = _axis_angle_rotation(axis, theta).transpose(0, 1)
+
+    rotated_pcd = rotate_pcd(pcd, matrix)
+
+    if batched:
+        [rotated_pcd] = unbatchify([rotated_pcd])
+
+    return rotated_pcd
+
+
 def scale_pcd(pcd: Tensor, scale_factor: Tensor) -> Tensor:
     """Scale a point cloud given the input scale factor.
 
@@ -626,7 +656,7 @@ def get_neighbouring_points(
         distances_sorted, indices_nn = torch.sort(distances, dim=-1)
         indices_first_nn = (
             indices_nn[:, :, 0]
-            .view(size_batch, num_pts_query, 1)
+            .reshape(size_batch, num_pts_query, 1)
             .repeat([1, 1, max_num_neighbours])
         )
         mask = distances_sorted == float("inf")
@@ -656,46 +686,57 @@ def voxelize_pcd(
     vmin: Optional[float] = None,
     vmax: Optional[float] = None,
 ) -> Tuple[Tensor, Tensor]:
-    """Voxelize point cloud.
+    """Voxelize point cloud(s).
 
-    The input cloud is voxelized according to the given resolution, i.e. the output voxel
-    grid will have shape (res, res, res). If vmin and vmax are passed, the function assumes
-    that the input cloud is already centered and scaled properly. Otherwise, the cloud is
-    centered and scaled in the unit sphere before computing the voxel grid.
+    Each input cloud is voxelized according to the given resolution,
+    i.e. the output voxel grid will have shape (res, res, res) for each cloud.
+    If vmin and vmax are passed, the function assumes that the input cloud
+    is already centered and scaled properly. Otherwise, the cloud is centered
+    and scaled in the unit sphere before computing the voxel grid.
 
     Args:
-        pcd: the input point cloud.
+        pcd: the input point cloud(s) with shape ([B,] NUM_POINTS, 3).
         res: the resolution for the voxel grid.
         vmin: The minimum value in x, y and z for the bounding box.
         vmax: The maximum value in x, y and z for the bounding box.
 
     Returns:
         A tuple containing:
-        - the voxel grid.
-        - the coordinates of the voxel grid.
+        - The voxel grid(s) with shape ([B,] res, res, res).
+        - The coordinates of the centroids of each cell of the voxel grid
+            with shape (res, res, res, 3). They are the same for each element
+            of the batch, so there's no need to return a batched version.
     """
+    batched, [pcd] = batchify([pcd], 3)
+
     if not vmin or not vmax:
         pcd = normalize_pcd_into_unit_sphere(pcd)
         vmin = -1.0
         vmax = 1.0
 
-    x = np.linspace(vmin, vmax, res)
-    x_grid, y_grid, z_grid = np.meshgrid(x, x, x)  # type: ignore
-    x_grid = x_grid.reshape((np.prod(x_grid.shape),))
-    y_grid = y_grid.reshape((np.prod(y_grid.shape),))
-    z_grid = z_grid.reshape((np.prod(z_grid.shape),))
-    grid_points_np = np.column_stack((x_grid, y_grid, z_grid))  # type: ignore
+    xyz_ranges = torch.tensor((vmax, vmax, vmax)) - torch.tensor((vmin, vmin, vmin))
+    xyz_ranges = xyz_ranges.to(pcd.device)
 
-    kdtree = KDTree(grid_points_np)
-    occupancies_np = np.zeros(len(grid_points_np), dtype=np.int8)
-    _, idx = kdtree.query(pcd.cpu().numpy())
-    occupancies_np[idx] = 1
+    leaf_sizes = xyz_ranges / res
+    ijk = ((pcd - vmin) / leaf_sizes.unsqueeze(0)).long()
+    ijk = torch.clip(ijk, 0, res - 1)
+    i, j, k = ijk[:, :, 0], ijk[:, :, 1], ijk[:, :, 2]
+    indices = i * (res**2) + j * res + k
 
-    occupancies = torch.tensor(occupancies_np, dtype=torch.float)
-    occupancies = torch.reshape(occupancies, shape=(res, res, res))
-    grid_points = torch.tensor(grid_points_np, dtype=torch.float)
+    occupancies = torch.zeros((pcd.shape[0], res**3), device=pcd.device)
+    ones = torch.ones_like(occupancies)
+    occupancies.scatter_(1, indices, ones)
+    occupancies = occupancies.reshape(pcd.shape[0], res, res, res)
 
-    return occupancies, grid_points
+    x = torch.arange(vmin, vmax, float(leaf_sizes[0]), device=pcd.device)
+    coords_x, coords_y, coords_z = torch.meshgrid(x, x, x, indexing="ij")
+    centroids_coords = torch.stack((coords_x, coords_y, coords_z), dim=-1).float()
+    centroids_coords += leaf_sizes[0] / 2
+
+    if batched:
+        [occupancies] = unbatchify([occupancies])
+
+    return occupancies, centroids_coords
 
 
 def knn(query: Tensor, target: Tensor, k: int) -> Tuple[Tensor, Tensor, Tensor]:
@@ -905,11 +946,12 @@ def compute_udf_from_pcd(
 
 
 def sample_pcds_from_udfs(
-    udfs_func: Callable[[Tensor], Tensor],
+    udfs_func: Callable[[Tensor, List[int]], Tensor],
     num_pcds: int,
     num_samples_forward: int,
     coords_range: Tuple[float, float],
     th_pred: float,
+    th_surf: float,
     num_out_points: int,
     num_steps_refinement: int,
 ) -> Tensor:
@@ -917,20 +959,26 @@ def sample_pcds_from_udfs(
 
     The function samples num_pcds point clouds in parallel, querying udfs_func with
     batch of coordinates. Thus, udfs_func must support batch forwards.
+    udfs_func is provided also with the list of indices of items that are currently computed:
+    e.g., if this function is used to sample a batch of 4 point clouds, initially udfs_func
+    will be called with indices [0, 1, 2, 3]. If after some iteration, no further sampling is
+    needed for clouds 1 and 3, udfs_func will be called with indices [0, 2].
     Each query is moved to the surface of its cloud according to the predicted UDF
     and to the gradient of the computed UDF. Additionally, this procedure is repeated
     iteratively to refine results, since predictions are supposed to be more accurate
     close to the surface.
 
     Args:
-        udfs_func: A function that will be queried with batch of coordinates (each item
-            of the batch corresponds to a different cloud). The function must return the
-            predicted UDF for each coordinate.
+        udfs_func: A function that will be queried with a batch of coordinates and with
+            a list of indices indicating to which item is associated each element of the
+            batch. The function must return the predicted UDF for each coordinate.
         num_samples_forward: The number of coordinates for each item of the batch when
             querying udfs_func.
         coords_range: The range for the queries coordinates.
-        th_pred: The threshold used to keep/discard queries. Only queries with predicted
-            UDF <= th_pred are kept.
+        th_pred: The threshold used to keep/discard queries after the first coarse prediction.
+            Only queries with predicted UDF <= th_pred are kept.
+        th_surf: The threshold used to keep/discard points after refinement. Only points
+            with refined UDF <= th_surf are kept.
         num_out_points: The number of points for the sampled clouds.
         num_steps_refinement: The number of refinement steps (at least 1).
 
@@ -953,22 +1001,22 @@ def sample_pcds_from_udfs(
         samples += coords_range[0]
 
         for refinement_step in range(num_steps_refinement):
-            samples = samples.view(len(remaining), -1, 3)
-            batch_indices = batch_indices.view(len(remaining), -1)
+            samples = samples.reshape(len(remaining), -1, 3)
+            batch_indices = batch_indices.reshape(len(remaining), -1)
 
             samples = samples.detach().clone()
             samples.requires_grad = True
 
-            preds = udfs_func(samples)
+            preds = udfs_func(samples, remaining)
             preds.sum().backward()  # type: ignore
 
             grads = cast(Tensor, samples.grad)
             directions = F.normalize(grads, dim=-1)
 
-            batch_indices = batch_indices.view(-1)
-            samples = samples.view(-1, 3)
-            preds = preds.view(-1)
-            directions = directions.view(-1, 3)
+            batch_indices = batch_indices.reshape(-1)
+            samples = samples.reshape(-1, 3)
+            preds = preds.reshape(-1)
+            directions = directions.reshape(-1, 3)
 
             if refinement_step == 0:
                 mask = preds <= th_pred
@@ -998,6 +1046,10 @@ def sample_pcds_from_udfs(
 
             samples = samples - directions * preds.unsqueeze(-1)
 
+        mask = preds <= th_surf
+        samples = samples[mask]
+        batch_indices = batch_indices[mask]
+
         for i in remaining:
             available_points = samples[batch_indices == i]
             needed_points = num_out_points - pcds[i].shape[0]
