@@ -51,10 +51,12 @@ class SIREN(nn.Module):
         """
         super().__init__()
 
+        assert num_hidden_layers > 0, "Number of hidden layers should be > 0."
+
         self.layers = nn.ModuleList()
 
         self.layers.append(nn.Sequential(nn.Linear(input_dim, hidden_dim), Sine()))
-        for _ in range(num_hidden_layers):
+        for _ in range(num_hidden_layers - 1):
             self.layers.append(nn.Sequential(nn.Linear(hidden_dim, hidden_dim), Sine()))
         self.layers.append(nn.Linear(hidden_dim, out_dim))
 
