@@ -1,22 +1,25 @@
+from typing import Tuple
+
 import pytorch3d.transforms.transform3d as pt3d  # type: ignore
 import torch
 from pytorch3d.transforms.rotation_conversions import _axis_angle_rotation  # type: ignore
 from torch import Tensor
 
-from pycarus.geometry.pcd import apply_affine_to_pcd, farthest_point_sampling, jitter_pcd
-from pycarus.geometry.pcd import normalize_pcd, random_drop_points, random_point_sampling
-from pycarus.geometry.pcd import rotate_pcd, scale_pcd, shuffle_pcd, translate_pcd
+from pycarus.geometry.pcd import apply_affine_to_pcd, farthest_point_sampling
+from pycarus.geometry.pcd import get_neighbouring_points, jitter_pcd, normalize_pcd_into_unit_cube
+from pycarus.geometry.pcd import normalize_pcd_into_unit_sphere, random_drop_points
+from pycarus.geometry.pcd import random_point_sampling, rotate_pcd, scale_pcd, shuffle_pcd
+from pycarus.geometry.pcd import translate_pcd
 
 
 class SamplePcd:
-    """Perform uniform sampling on point cloud using farthest point sampling.
-
-    Args:
-        num_points: Number of point to sample.
-        algorithm: the algorithm to use for the subsampling one in ["fps", "random"].
-    """
-
     def __init__(self, num_points: int, algorithm: str) -> None:
+        """Subsample the input point cloud(s) using the given algorithm.
+
+        Args:
+            num_points: Number of point to sample.
+            algorithm: the algorithm to use for the subsampling between ["fps", "random"].
+        """
         self.num_points = num_points
         self.algorithm = algorithm
 
@@ -47,7 +50,22 @@ class SamplePcd:
         return format_string
 
 
-class NormalizePcd:
+class NormalizePcdIntoUnitSphere:
+    """Normalize a given point cloud into the unit sphere."""
+
+    def __call__(self, pcd: Tensor) -> Tensor:
+        """Normalize a point cloud.
+
+        Args:
+            pcd: The input point cloud with shape (NUM_POINTS, D).
+
+        Returns:
+            The normalized point cloud.
+        """
+        return normalize_pcd_into_unit_sphere(pcd)
+
+
+class NormalizePcdIntoUnitCube:
     """Normalize a given point cloud into the unit cube."""
 
     def __call__(self, pcd: Tensor) -> Tensor:
@@ -59,12 +77,10 @@ class NormalizePcd:
         Returns:
             The normalized point cloud.
         """
-        return normalize_pcd(pcd)
+        return normalize_pcd_into_unit_cube(pcd)
 
 
 class AffinePcd(pt3d.Transform3d):
-    """Transform to apply an affine transformation, i.e. rotation plus a translation."""
-
     def __init__(self, matrix: Tensor) -> None:
         """Transform to apply an affine transformation, i.e. rotation plus a translation.
 
@@ -79,8 +95,6 @@ class AffinePcd(pt3d.Transform3d):
 
 
 class RotatePcd(pt3d.Rotate):
-    """Transform to rotate a point cloud."""
-
     def __init__(self, rotation: Tensor) -> None:
         """Transform to rotate a point cloud.
 
@@ -95,8 +109,6 @@ class RotatePcd(pt3d.Rotate):
 
 
 class RandomRotatePcdAroundAxis:
-    """Rotate a point cloud around one axis by a random angle sampled within a range."""
-
     def __init__(self, axis: str, min_angle: float = 0, max_angle: float = 180):
         """Rotate a point cloud around one axis by a random angle sampled within a range.
 
@@ -130,12 +142,11 @@ class RandomRotatePcdAroundAxis:
             The rotation matrix with shape (3, 3).
         """
         theta = torch.FloatTensor(1).uniform_(min_angle, max_angle)
-        matrix_col_major = _axis_angle_rotation(axis, theta)
-        # The convention used in this framework is to operate on points as row vectors.
-        # The rotation matrix returned from _axis_angle_rotation is for transforming column vectors.
-        # Therefore we transpose this matrix.
-        matrix_row_major = matrix_col_major.transpose(1, 2)[0]
-        return matrix_row_major
+        matrix = _axis_angle_rotation(axis, theta)[0]
+        # _axis_angle_rotation returns a rotation matrix thought to operate
+        # on column vectors, since we use row vectors we need to transpose
+        matrix = matrix.transpose(0, 1)
+        return matrix
 
     def __call__(self, pcd: Tensor) -> Tensor:
         self.last_matrix = self._get_params(self.min_angle, self.max_angle, self.axis)
@@ -149,8 +160,6 @@ class RandomRotatePcdAroundAxis:
 
 
 class TranslatePcd(pt3d.Translate):
-    """Transform to translate a point cloud."""
-
     def __init__(self, translation: Tensor) -> None:
         """Transform to translate a point cloud.
 
@@ -166,8 +175,6 @@ class TranslatePcd(pt3d.Translate):
 
 
 class RandomTranslatePcd:
-    """Transform to translate randomly a point cloud."""
-
     def __init__(self, min_tr: float = -0.5, max_tr: float = 0.5) -> None:
         """Transform to translate randomly a point cloud.
 
@@ -184,8 +191,6 @@ class RandomTranslatePcd:
 
 
 class ScalePcd(pt3d.Scale):
-    """Transform to scale a point cloud."""
-
     def __init__(self, scale_factor: Tensor) -> None:
         """Transform to scale a point cloud.
 
@@ -201,8 +206,6 @@ class ScalePcd(pt3d.Scale):
 
 
 class RandomScalePcd:
-    """Transform to scale randomly a point cloud."""
-
     def __init__(self, min_scale: float = 0.5, max_scale: float = 1.5) -> None:
         """Transform to scale randomly a point cloud.
 
@@ -219,8 +222,6 @@ class RandomScalePcd:
 
 
 class JitterPcd:
-    """Jitter point cloud by adding random noise"""
-
     def __init__(self, sigma: float, clip: float) -> None:
         """Jitter point cloud by adding random noise.
 
@@ -236,8 +237,6 @@ class JitterPcd:
 
 
 class RandomDropPcd:
-    """Randomly remove points from a point cloud."""
-
     def __init__(self, min_percentage: float, max_percentage: float) -> None:
         """Remove random percentage of points from a point cloud. The number of points is uniformly
         sampled in the range min_percentage - max_percentage.
@@ -289,3 +288,41 @@ class ShufflePcd:
 
     def __repr__(self) -> str:
         return f"{self.__class__.__name__}"
+
+
+class Patcher:
+    """Extract patches from a point cloud."""
+
+    def __init__(self, num_points_per_patch: int, radius: float) -> None:
+        """Create a Patcher.
+
+        Args:
+            num_points_per_patch: The number of points for each patch.
+            radius: Radius for the for radius based search.
+        """
+        self.num_points_per_patch = num_points_per_patch
+        self.radius = radius
+
+    def __call__(self, pcd: Tensor, kpts: Tensor) -> Tuple[Tensor, Tensor]:
+        """Get a predefined number of patches from given point cloud(s).
+        Each patch will have the same number of points with coordinates
+        centered on the patch keypoint.
+
+        Args:
+            pcd: The input point cloud(s) with shape ([B,] NUM_POINTS, D).
+            kpts: The selected keypoints, one for each patch, with shape ([B,] NUM_PATCHES, 3).
+
+        Returns:
+            - The global indices with shape ([B,] NUM_PATCHES, NUM_POINTS_PER_PATCH).
+            - The patches with shape ([B,] NUM_PATCHES, NUM_POINTS_PER_PATCH, D).
+        """
+        indices, patches = get_neighbouring_points(
+            pcd,
+            kpts,
+            self.num_points_per_patch,
+            self.radius,
+        )
+
+        patches -= kpts.unsqueeze(-2)
+
+        return indices, patches
